#include <msp/core/refptr.h>
#include <msp/io/file.h>
#include <msp/strings/format.h>
#include "bmploader.h"
#include "imageloader.h"
#ifdef WITH_LIBPNG
#include "png/pngloader.h"
#endif
#ifdef WITH_LIBJPEG
#include "jpeg/jpegloader.h"
#endif
#ifdef WITH_DEVIL
#include "devil/devilloader.h"
#endif
#ifdef WITH_QUARTZ
#include "quartz/quartzloader.h"
#endif

using namespace std;

namespace Msp {
namespace Graphics {

ImageLoader::ImageLoader():
	source(0),
	state(INITIAL)
{ }

ImageLoader::~ImageLoader()
{
	delete source;
}

ImageLoader *ImageLoader::open_file(const string &fn)
{
	try
	{
		RefPtr<IO::BufferedFile> file = new IO::BufferedFile(fn);
		ImageLoader *loader = open_io(*file);
		loader->source = file.release();
		return loader;
	}
	catch(const unsupported_image_format &e)
	{
		throw unsupported_image_format(format("%s: %s", fn, e.what()));
	}
}

ImageLoader *ImageLoader::open_io(IO::Seekable &io)
{
	Registry &registry = get_registry();
	if(registry.changed)
	{
		registry.changed = false;
		registry.loaders.sort(signature_size_compare);
	}

	if(registry.loaders.empty())
		throw unsupported_image_format("no loaders");

	vector<char> sig_buf(registry.loaders.back()->get_signature_size());
	unsigned sig_len = io.read(&sig_buf[0], sig_buf.size());
	string signature(sig_buf.begin(), sig_buf.end());

	ImageLoader *loader = 0;
	for(list<RegisterBase *>::const_iterator i=registry.loaders.begin(); (!loader && i!=registry.loaders.end()); ++i)
		if((*i)->detect(signature))
			loader = (*i)->create(io);

	io.seek(0, IO::S_BEG);

	if(!loader)
	{
		string sig_hex;
		for(unsigned i=0; i<sig_len; ++i)
		{
			if(i)
				sig_hex += ' ';
			sig_hex += format("%02X", static_cast<unsigned char>(sig_buf[i]));
		}
		throw unsupported_image_format(sig_hex);
	}

	return loader;
}

void ImageLoader::load(Image::Data &data)
{
	if(state>=FINISHED)
		throw logic_error("already loaded");

	if(state<HEADERS_LOADED)
		load_headers_(data);
	if(!data.pixels)
		data.pixels = data.owned_pixels = new char[data.stride*data.height];
	load_pixels_(data);
	state = FINISHED;
}

void ImageLoader::load_headers(Image::Data &data)
{
	if(state>=HEADERS_LOADED)
		throw logic_error("headers already loaded");

	load_headers_(data);
	state = HEADERS_LOADED;
}

ImageLoader::Registry &ImageLoader::get_registry()
{
	static Registry registry;
	static bool initialized = false;
	if(!initialized)
	{
		initialized = true;
		register_loader<BmpLoader>();
#ifdef WITH_LIBPNG
		register_loader<PngLoader>();
#endif
#ifdef WITH_LIBJPEG
		register_loader<JpegLoader>();
#endif
#ifdef WITH_DEVIL
		register_loader<DevilLoader>();
#endif
#ifdef WITH_QUARTZ
		register_loader<QuartzLoader>();
#endif
	}
	return registry;
}

bool ImageLoader::signature_size_compare(RegisterBase *r1, RegisterBase *r2)
{
	return r1->get_signature_size()<r2->get_signature_size();
}


ImageLoader::Registry::Registry():
	changed(false)
{ }

ImageLoader::Registry::~Registry()
{
	for(list<RegisterBase *>::iterator i=loaders.begin(); i!=loaders.end(); ++i)
		delete *i;
}

} // namespace Graphics
} // namespace Msp
