#include <msp/core/application.h>
#include "display.h"
#include "window.h"
#include "window_private.h"

using namespace std;

namespace Msp {
namespace Graphics {

WindowOptions::WindowOptions():
	x(0),
	y(0),
	user_position(false),
	width(640),
	height(480),
	fullscreen(false),
	fullscreen_monitor(0),
	fullscreen_exclusive(true),
	resizable(false)
{ }


Window::Window(Display &dpy, unsigned w, unsigned h, bool fs):
	display(dpy)
{
	options.width = w;
	options.height = h;
	options.fullscreen = fs;

	init();
}

Window::Window(Display &dpy, const WindowOptions &opts):
	display(dpy),
	options(opts)
{
	init();
}

void Window::init()
{
	visible = false;
	kbd_autorepeat = true;
	touch_input = false;
	resizing = false;
	moving = false;
	priv = new Private;

	if(options.fullscreen && !options.fullscreen_monitor)
		options.fullscreen_monitor = display.get_desktop_mode().monitor;

	platform_init();

	display.add_window(*this);
	display.check_error();

	const string &app_name = Application::get_name();
	if(!app_name.empty())
		set_title(app_name);
}

Window::~Window()
{
	platform_cleanup();

	display.remove_window(*this);

	if(options.fullscreen && visible)
		display.restore_mode();

	delete priv;
}

void Window::reconfigure(const WindowOptions &opts)
{
	bool fullscreen_changed = (opts.fullscreen!=options.fullscreen);
	if(opts.width!=options.width || opts.height!=options.height)
		resizing = true;
	if(opts.x!=options.x || opts.y!=options.y)
		moving = true;

	options = opts;

	if(visible)
	{
		if(options.fullscreen)
			set_fullscreen_mode();
		else if(fullscreen_changed)
			display.restore_mode();
	}

	platform_reconfigure(fullscreen_changed);
}

void Window::set_fullscreen_mode()
{
	if(!options.fullscreen_monitor)
		options.fullscreen_monitor = display.get_desktop_mode().monitor;
	VideoMode mode(options.width, options.height);
	mode.monitor = options.fullscreen_monitor;
	mode.rotation = mode.monitor->desktop_settings.rotation;
	if(mode.rotation==ROTATE_LEFT || mode.rotation==ROTATE_RIGHT)
		swap(mode.width, mode.height);
	display.set_mode(mode, options.fullscreen_exclusive);
}

void Window::set_keyboard_autorepeat(bool r)
{
	kbd_autorepeat = r;
}

void Window::set_touch_input(bool t)
{
	touch_input = t;
	platform_set_touch_input();
}

void Window::show()
{
	platform_show();
	visible = true;

	if(options.fullscreen)
	{
		set_fullscreen_mode();
		warp_pointer(options.width/2, options.height/2);
	}
}

void Window::hide()
{
	platform_hide();
	visible = false;

	if(options.fullscreen)
		display.restore_mode();
}

} // namespace Graphics
} // namespace Msp
